#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import datetime
from typing import Iterable, List

import pytest

from pydifact.segmentcollection import (
    Interchange, Message, RawSegmentCollection, SegmentCollection)
from pydifact.segments import Segment
from pydifact.api import EDISyntaxError


def test_from_file():
    with pytest.raises(FileNotFoundError):
        SegmentCollection.from_file("/no/such/file")


def test_create_with_segments():
    collection = SegmentCollection.from_segments([Segment("36CF")])
    assert [Segment("36CF")] == collection.segments


def test_get_segments():
    collection = SegmentCollection.from_segments(
        [Segment("36CF", 1), Segment("CPD"), Segment("36CF", 2)]
    )
    segments = list(collection.get_segments("36CF"))
    assert [Segment("36CF", 1), Segment("36CF", 2)] == segments


def test_get_segments_doesnt_exist():
    collection = SegmentCollection()
    segments = list(collection.get_segments("36CF"))
    assert [] == segments


def test_get_segments_w_predicate():
    collection = SegmentCollection.from_segments(
        [
            Segment("A", "1", "a"),
            Segment("A", "2", "b"),
            Segment("A", "1", "c"),
        ]
    )
    segments = collection.get_segments("A", lambda x: x[0] == "1")
    assert [
        Segment("A", "1", "a"),
        Segment("A", "1", "c"),
    ] == list(segments)


def test_get_segment():
    collection = SegmentCollection.from_segments(
        [Segment("36CF", 1), Segment("36CF", 2)]
    )
    segment = collection.get_segment("36CF")
    assert Segment("36CF", 1) == segment


def test_get_segment_w_predicate():
    collection = SegmentCollection.from_segments(
        [Segment("36CF", "1"), Segment("36CF", "2")]
    )
    segment = collection.get_segment("36CF", lambda x: x[0] == "2")
    assert segment == Segment("36CF", "2")


def test_split_by():
    def _serialize(
            collections: Iterable[RawSegmentCollection]
    ) -> List[List[str]]:
        lst = []
        global_lst = []
        for collection in collections:
            if lst:
                global_lst.append(lst)
                lst = []
            for segment in collection.segments:
                lst.append(segment.tag)
        if lst:
            global_lst.append(lst)
        return global_lst

    assert (
        _serialize(RawSegmentCollection.from_segments([]).split_by('A'))
        ==
        []
    )
    collection = RawSegmentCollection.from_segments(
        Segment(i) for i in ["A", "B", "A", "A", "B", "D"]
    )
    assert _serialize(collection.split_by('Z')) == []
    assert (
        _serialize(collection.split_by('A'))
        ==
        [['A', 'B'], ['A'], ['A', 'B', 'D']]
    )
    assert (
        _serialize(collection.split_by('A'))
        ==
        [['A', 'B'], ['A'], ['A', 'B', 'D']]
    )


def test_str_serialize():
    collection = SegmentCollection.from_segments(
        [Segment("36CF", "1"), Segment("36CF", "2")]
    )
    string = str(collection)
    assert "36CF+1'36CF+2'" == string


def test_get_segment_doesnt_exist():
    collection = SegmentCollection()
    segment = collection.get_segment("36CF")
    assert segment is None


def test_empty_segment():
    m = SegmentCollection()
    m.add_segment(Segment("", []))
    assert m


def test_malformed_tag1():
    with pytest.raises(EDISyntaxError):
        SegmentCollection.from_str("IMD+F++:::This is 'a :malformed string'")


def test_malformed_tag2():
    with pytest.raises(EDISyntaxError):
        SegmentCollection.from_str("IMD+F++:::This is '? :malformed string'")


def test_malformed_tag3():
    with pytest.raises(EDISyntaxError):
        SegmentCollection.from_str("IMD+F++:::This is '?? :malformed string'")


def test_malformed_tag4():
    with pytest.raises(EDISyntaxError):
        SegmentCollection.from_str("IMD+F++:::This is '??:malformed string'")


def test_malformed_tag5():
    with pytest.raises(EDISyntaxError):
        SegmentCollection.from_str("IMD+F++:::This is '-:malformed string'")


@pytest.fixture
def interchange():
    return Interchange(
        sender="1234",
        recipient="3333",
        timestamp=datetime.datetime(2020, 1, 2, 22, 12),
        control_reference="42",
        syntax_identifier=("UNOC", 1),
    )


@pytest.fixture
def message():
    return Message(
        reference_number="42z42",
        identifier=("PAORES", 93, 1, "IA"),
    )


def test_empty_interchange(interchange):
    assert str(interchange) == ("UNB+UNOC:1+1234+3333+200102:2212+42'" "UNZ+0+42'")


def test_empty_interchange_w_extra_header(interchange):
    i = Interchange(
        sender="1234",
        recipient="3333",
        timestamp=datetime.datetime(2020, 1, 2, 22, 12),
        control_reference="42",
        syntax_identifier=("UNOC", 1),
        extra_header_elements=[["66", "2"], "ZZ"],
    )

    assert str(i) == ("UNB+UNOC:1+1234+3333+200102:2212+42+66:2+ZZ'" "UNZ+0+42'")


def test_empty_interchange_from_str():
    i = Interchange.from_str("UNB+UNOC:1+1234+3333+200102:2212+42'" "UNZ+0+42'")
    assert str(i) == ("UNB+UNOC:1+1234+3333+200102:2212+42'" "UNZ+0+42'")


def test_empty_interchange_w_una():
    i = Interchange.from_segments(
        [
            Segment("UNA", ":+,? '"),
            Segment("UNB", ["UNOC", "1"], "1234", "3333", ["200102", "2212"], "42"),
            Segment("UNZ", "0", "42"),
        ]
    )
    assert str(i) == ("UNA:+,? '" "UNB+UNOC:1+1234+3333+200102:2212+42'" "UNZ+0+42'")


def test_interchange_messages(interchange, message):
    assert len(list(interchange.get_messages())) == 0

    interchange.add_message(message)

    assert len(list(interchange.get_messages())) == 1

    assert str(interchange) == (
        "UNB+UNOC:1+1234+3333+200102:2212+42'"
        "UNH+42z42+PAORES:93:1:IA'"
        "UNT+42z42+0'"
        "UNZ+2+42'"
    )


def test_interchange_messages_from_str():
    i = Interchange.from_str(
        "UNB+UNOC:1+1234+3333+200102:2212+42'"
        "UNH+42z42+PAORES:93:1:IA'"
        "UNT+42z42+0'"
        "UNZ+2+42'"
    )
    assert str(i) == (
        "UNB+UNOC:1+1234+3333+200102:2212+42'"
        "UNH+42z42+PAORES:93:1:IA'"
        "UNT+42z42+0'"
        "UNZ+2+42'"
    )


def test_faulty_interchange_messages():
    i = Interchange.from_str(
        "UNB+UNOC:1+1234+3333+200102:2212+42'"
        "UNH+42z42+PAORES:93:1:IA'"
        "UNH+42z42+PAORES:93:1:IA'"
        "UNZ+2+42'"
    )

    with pytest.raises(SyntaxError):
        list(i.get_messages())

    i = Interchange.from_str(
        "UNB+UNOC:1+1234+3333+200102:2212+42'" "UNT+42z42+0'" "UNZ+2+42'"
    )

    with pytest.raises(SyntaxError):
        list(i.get_messages())


def test_empty_message(message):
    assert str(message) == ("UNH+42z42+PAORES:93:1:IA'" "UNT+42z42+0'")



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
from pydifact.segmentcollection import SegmentCollection


# This only a performance benchmark, no real test.
def performance_test_huge_message():
    """Performance test parsing a huge message"""
    collection = SegmentCollection.from_file("tests/data/huge_file2.edi")
    assert collection


if __name__ == "__main__":
    # just call performance test function for profiling purposes
    # use this file with:
    #    $ python -m cProfile -s time tests/test_huge_message.py
    performance_test_huge_message()



# Pydifact - a python edifact library
#
# Copyright (c) 2019 Christian González
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from enum import Enum


class Token:
    """Represents a block of characters in the message.

    This could be content, a data separator (usually +),
    a component data separator (usually :), or a segment terminator (usually ').
    """

    class Type(Enum):
        CTRL_CHARS = 10  # ASCII string holding the control chars
        CONTENT = 11
        COMPONENT_SEPARATOR = 12  # default :
        DATA_SEPARATOR = 13  # default +
        TERMINATOR = 14  # default '

    def __init__(self, token_type: Type, value: str):
        """Creates a Token with a type and a value"""
        assert type(token_type) == Token.Type

        self.type = token_type
        self.value = value

    def __str__(self):
        return "{name} Token: '{value}'".format(name=self.type.name, value=self.value)

    def __repr__(self):
        return "<{name} Token object '{value}' at {address}>".format(
            name=self.type.name, value=self.value, address=hex(id(self))
        )

    def __eq__(self, other) -> bool:
        return self.type == other.type and self.value == other.value



from pydifact.segments import Segment
from pydifact.syntax.common import UNASegment

default_characters = ":+,? '"


def test_empty_UNASegment():
    s = UNASegment()
    assert s == Segment("UNA", default_characters)


# TODO: add more checks



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import pytest

from pydifact.token import Token
from pydifact.tokenizer import Tokenizer

from pydifact.control import Characters


@pytest.fixture
def tokenizer():
    return Tokenizer()


def _assert_tokens(
    collection: str, expected: list = None, error_message: str = None
) -> None:
    """Helper function to accelerate tokenizer testing."""

    if expected is None:
        expected = []
    tokens = list(Tokenizer().get_tokens(collection))
    if error_message:
        assert expected == tokens, error_message
    else:
        assert expected == tokens


def test_basic():
    _assert_tokens(
        "RFF+PD:50515'",
        [
            Token(Token.Type.CONTENT, "RFF"),
            Token(Token.Type.DATA_SEPARATOR, "+"),
            Token(Token.Type.CONTENT, "PD"),
            Token(Token.Type.COMPONENT_SEPARATOR, ":"),
            Token(Token.Type.CONTENT, "50515"),
            Token(Token.Type.TERMINATOR, "'"),
        ],
    )


def test_escape():
    _assert_tokens(
        "RFF+PD?:5'",
        [
            Token(Token.Type.CONTENT, "RFF"),
            Token(Token.Type.DATA_SEPARATOR, "+"),
            Token(Token.Type.CONTENT, "PD:5"),
            Token(Token.Type.TERMINATOR, "'"),
        ],
    )


def test_double_escape():
    _assert_tokens(
        "RFF+PD??:5'",
        [
            Token(Token.Type.CONTENT, "RFF"),
            Token(Token.Type.DATA_SEPARATOR, "+"),
            Token(Token.Type.CONTENT, "PD?"),
            Token(Token.Type.COMPONENT_SEPARATOR, ":"),
            Token(Token.Type.CONTENT, "5"),
            Token(Token.Type.TERMINATOR, "'"),
        ],
    )


def test_triple_escape():
    _assert_tokens(
        "RFF+PD???:5'",
        [
            Token(Token.Type.CONTENT, "RFF"),
            Token(Token.Type.DATA_SEPARATOR, "+"),
            Token(Token.Type.CONTENT, "PD?:5"),
            Token(Token.Type.TERMINATOR, "'"),
        ],
    )


def test_quadruple_escape():
    _assert_tokens(
        "RFF+PD????:5'",
        [
            Token(Token.Type.CONTENT, "RFF"),
            Token(Token.Type.DATA_SEPARATOR, "+"),
            Token(Token.Type.CONTENT, "PD??"),
            Token(Token.Type.COMPONENT_SEPARATOR, ":"),
            Token(Token.Type.CONTENT, "5"),
            Token(Token.Type.TERMINATOR, "'"),
        ],
    )


def test_starts_with_escape():
    _assert_tokens(
        "DTM+?+0'",
        [
            Token(Token.Type.CONTENT, "DTM"),
            Token(Token.Type.DATA_SEPARATOR, "+"),
            Token(Token.Type.CONTENT, "+0"),
            Token(Token.Type.TERMINATOR, "'"),
        ],
    )


# This tests check if line break combinations (CR/LF) after a segment terminator are correctly ignored.


@pytest.fixture
def expected_crlf():
    return [
        Token(Token.Type.CONTENT, "RFF"),
        Token(Token.Type.COMPONENT_SEPARATOR, ":"),
        Token(Token.Type.CONTENT, "5"),
        Token(Token.Type.TERMINATOR, "'"),
        Token(Token.Type.CONTENT, "DEF"),
        Token(Token.Type.COMPONENT_SEPARATOR, ":"),
        Token(Token.Type.CONTENT, "6"),
        Token(Token.Type.TERMINATOR, "'"),
    ]


def test_ignore_lf(expected_crlf):
    _assert_tokens("RFF:5'\nDEF:6'", expected_crlf)


def test_ignore_crlf(expected_crlf):
    _assert_tokens("RFF:5'\r\nDEF:6'", expected_crlf)


def test_ignore_cr(expected_crlf):
    _assert_tokens("RFF:5'\rDEF:6'", expected_crlf)


def test_ignore_lfcr(expected_crlf):
    _assert_tokens("RFF:5'\n\rDEF:6'", expected_crlf)


def test_ignore_lfcr_combined(expected_crlf):
    _assert_tokens("RFF:5'\n\r\n\r\nDEF:6'\n\r", expected_crlf)


def test_ignore_whitespace(expected_crlf):
    _assert_tokens("RFF:5'   \nDEF:6'", expected_crlf)


def test_ignore_long_whitespace(expected_crlf):
    _assert_tokens("RFF:5'               \nDEF:6'", expected_crlf)


def test_no_terminator():
    with pytest.raises(RuntimeError):
        list(Tokenizer().get_tokens("TEST", Characters()))
        pytest.fail("Unexpected end of EDI message")



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os

from pydifact.segmentcollection import SegmentCollection
from pydifact.segments import Segment

path = os.path.dirname(os.path.realpath(__file__)) + "/data"


def test_wikipedia_file():
    message = SegmentCollection.from_file("{}/wikipedia.edi".format(path))
    # make some checks
    assert message.get_segment("UNB") == Segment(
        "UNB", ["IATB", "1"], "6XPPC", "LHPPC", ["940101", "0950"], "1"
    )
    assert message.get_segment("IFT") == Segment("IFT", "3", "XYZCOMPANY AVAILABILITY")
    assert message.get_segment("TVL") == Segment(
        "TVL", ["240493", "1000", "", "1220"], "FRA", "JFK", "DL", "400", "C"
    )


def test_invoice_file():
    message = SegmentCollection.from_file("{}/invoice1.edi".format(path))
    # make some checks
    assert message.get_segment("UNB") == Segment(
        "UNB",
        ["UNOA", "1"],
        "01010000253001",
        "O0013000093SCHA-Z59",
        ["991006", "1902"],
        "PAYO0012101221",
    )
    assert message.get_segment("DTM") == Segment("DTM", ["137", "199910060000", "102"])
    assert message.get_segment("NAD") == Segment(
        "NAD", "BT", ["VAUXHALL MOTORS LTD", "", "91"]
    )
    assert message.get_segment("RFF") == Segment("RFF", ["VA", "382324067"])


# def test_order_file():
#     _test_file_read("{}/order.edi".format(path))
#
#
# def test_patient1_file():
#     _test_file_read("{}/patient1.edi".format(path))



# Pydifact - a python edifact library
#
# Copyright (c) 2019 Christian González
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Union, List

from pydifact.api import EDISyntaxError, PluginMount
from pydifact.control import Characters


class SegmentProvider(metaclass=PluginMount):
    """This is a plugin mount point for Segment plugins which represent a certain EDIFACT Segment.

    Classes implementing this PluginMount should provide the following attributes:
    """

    def __str__(self):
        """Returns the user readable text representation of this segment."""

    def validate(self) -> bool:
        """Validates the Segment."""


class Segment(SegmentProvider):
    """Represents a low-level segment of an EDI interchange.

    This class is used internally. read-world implementations of specialized should subclass Segment and provide
    the `tag` and `validate` attributes.
    """

    # tag is not a class attribute in this case, as each Segment instance could have another tag.
    __omitted__ = True

    def __init__(self, tag: str, *elements: Union[str, List[str]]):
        """Create a new Segment instance.

        :param str tag: The code/tag of the segment. Must not be empty.
        :param list elements: The data elements for this segment, as (possibly empty) list.

        """
        self.tag = tag

        # The data elements for this segment.
        # this is converted to a list (due to the fact that python creates a tuple
        # when passing a variable arguments list to a method)
        self.elements = list(elements)

    def __str__(self) -> str:
        """Returns the Segment in Python list printout"""
        return "'{tag}' EDI segment: {elements}".format(
            tag=self.tag, elements=str(self.elements)
        )

    def __repr__(self) -> str:
        return "{} segment: {}".format(self.tag, str(self.elements))

    def __eq__(self, other) -> bool:
        # FIXME the other way round too? isinstance(other, type(self))?
        return (
            isinstance(self, type(other))
            and self.tag == other.tag
            and list(self.elements) == list(other.elements)
        )

    def __getitem__(self, key):
        return self.elements[key]

    def __setitem__(self, key, value):
        self.elements[key] = value

    def validate(self) -> bool:
        """
        Segment validation.

        The Segment class is part of the lower level interfaces of pydifact.
        So it assumes that the given parameters are correct, there is no validation done here.
        However, in segments derived from this class, there should be validation.

        :return: bool True if given tag and elements are a valid EDIFACT segment, False if not.
        """
        # FIXME: there should be a way of returning an error message - WHICH kind of validation failed.

        if not self.tag:
            return False
        return True


class SegmentFactory:
    """Factory for producing segments."""

    characters = None

    @staticmethod
    def create_segment(
        name: str, *elements: Union[str, List[str]], validate: bool = True
    ) -> Segment:
        """Create a new instance of the relevant class type.

        :param name: The name of the segment
        :param elements: The data elements for this segment
        :param validate: bool if True, the created segment is validated before return
        """
        if not SegmentFactory.characters:
            SegmentFactory.characters = Characters()

        # Basic segment type validation is done here.
        # The more special validation must be done in the corresponding Segment

        if not name:
            raise EDISyntaxError("The tag of a segment must not be empty.")

        if type(name) != str:
            raise EDISyntaxError(
                "The tag name of a segment must be a str, but is a {}: {}".format(
                    type(name), name
                )
            )

        if not name.isalnum():
            raise EDISyntaxError(
                "Tag '{}': A tag name must only contain alphanumeric characters.".format(
                    name
                )
            )

        for Plugin in SegmentProvider.plugins:
            if Plugin().tag == name:
                s = Plugin(name, *elements)
        else:
            # we don't support this kind of EDIFACT segment (yet), so
            # just create a generic Segment()
            s = Segment(name, *elements)

        if validate:
            if not s.validate():
                raise EDISyntaxError(
                    "could not create '{}' Segment. Validation failed.".format(name)
                )

        # FIXME: characters is not used!
        return Segment(name, *elements)



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import pytest

from pydifact.segments import Segment


elements = ["field1", ["field2", "extra"], "stuff"]


def test_get_segment_code():
    segment = Segment("OMD")
    assert segment.tag == "OMD"


def test_all_elements():
    segment = Segment("OMD", *elements)
    assert segment.elements == elements


def test_get_single_element():
    segment = Segment("OMD", *elements)
    assert segment.elements[0] == "field1"


def test_get_list_element():
    segment = Segment("OMD", *elements)
    assert segment.elements[1] == ["field2", "extra"]


def test_get_non_existing_element():
    segment = Segment("OMD", *elements)
    with pytest.raises(IndexError):
        segment.elements[7]



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from pydifact.control import Characters


def test_with_separator_identity():
    one = Characters()
    other = Characters()
    # a copy of a characters object must be equal, but not the same
    assert one == other, 'Objects differ: "{}", "{}"'.format(one, other)
    assert one is not other


def test_cc_assigning():
    one = Characters()
    one.component_separator = "x"
    assert one.component_separator == "x"
    assert one == "x+,? '"


def test_wrong_cc_assigning():
    with pytest.raises(ValueError):
        Characters().with_control_character("component_separator", "xd")

    with pytest.raises(AttributeError):
        Characters().with_control_character("notexisting", ":")



# Pydifact - a python edifact library
#
# Copyright (c) 2019 Christian González
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import List, Optional

from pydifact.control.characters import Characters
import re

from pydifact.segments import Segment


class Serializer:
    """Serialize a bunch of segments into an EDI message string."""

    def __init__(self, characters: Characters = None):
        super().__init__()
        if characters is None:
            characters = Characters()

        self.characters = characters
        self.replace_map = characters.escaped_syntax_dic

        # Thanks to "Bor González Usach" for this wonderful piece of code:
        # https://gist.github.com/bgusach/a967e0587d6e01e889fd1d776c5f3729
        substrs = sorted(self.replace_map, key=len, reverse=True)
        self.regexp = re.compile("|".join(map(re.escape, substrs)))

    def serialize(
        self, segments: List[Segment], with_una_header: bool = True, break_lines=False
    ) -> str:
        """Serialize all the passed segments.

        :param segments: A list of segments to serialize
        :param with_una_header: includes/adds an UNA header if set to True (=default)
            If the segments list contains a UNA header, it is taken, else the default character set
            is created.
        :param break_lines: if True, insert line break after each segment terminator.
        """
        collection_parts = []

        # first, check if UNA header is wanted.
        if with_una_header:
            if not segments:
                return self.characters.service_string_advice
            else:
                collection_parts += self.characters.service_string_advice

        else:
            # no una header wanted!
            if not segments:
                return ""

        # iter through all segments
        for segment in segments:
            # skip the UNA segment as we already have written it if requested
            if segment.tag == "UNA":
                continue
            collection_parts += segment.tag
            for element in segment.elements:
                collection_parts += self.characters.data_separator
                if type(element) == list:
                    element = (self.escape(subelement) for subelement in element)
                    collection_parts += self.characters.component_separator.join(
                        element
                    )

                else:
                    collection_parts += self.escape(element)

            collection_parts += self.characters.segment_terminator
            if break_lines:
                collection_parts += "\n"

        collection = "".join(collection_parts)
        return collection

    def escape(self, string: Optional[str]) -> str:
        """Escapes control characters.

        :param string: The string to be escaped
        """

        if string is None:
            return ""
        assert type(string) == str, "%s is not a str, it is %s" % (string, type(string))

        return self.regexp.sub(lambda match: self.replace_map[match.group(0)], string)



# Pydifact - a python edifact library
#
# Copyright (c) 2019 Christian González
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from typing import Optional, Generator, Any

from pydifact.tokenizer import Tokenizer
from pydifact.token import Token
from pydifact.segments import Segment, SegmentFactory
from pydifact.control import Characters


class Parser:
    """Parse EDI messages into a list of segments."""

    def __init__(self, factory: SegmentFactory = None):
        if factory is None:
            factory = SegmentFactory()

        self.factory = factory
        self.characters = Characters()

    def parse(
        self, message: str, characters: Characters = None
    ) -> Generator[Segment, Any, None]:
        """Parse the message into a list of segments.

        :param characters: the control characters to use, if there is no
                UNA segment present
        :param message: The EDI message
        :rtype:
        """

        # If there is a UNA, take the following 6 characters
        # unconditionally, save them, strip them, and make control Characters()
        # for further parsing
        if message[0:3] == "UNA":
            self.characters = Characters.from_str("UNA" + message[3:9])

            # remove the UNA segment from the string
            message = message[9:].lstrip("\r\n")

        else:
            # if no UNA header present, use default control characters
            if characters is not None:
                self.characters = characters

        tokenizer = Tokenizer()
        return self.convert_tokens_to_segments(
            tokenizer.get_tokens(message, self.characters), self.characters
        )

    @staticmethod
    def get_control_characters(
        message: str, characters: Characters = None
    ) -> Optional[Characters]:
        """Read the UNA segment from the passed string and extract/store the control characters from it.

        :param message: a valid EDI message string, or UNA segment string,
                        to extract the control characters from.
        :param characters: the control characters to use, if none found in
                           the message. Default: ":+,? '"
        :return: the control characters
        """

        if not characters:
            characters = Characters()

        # First, try to read the UNA segment ("Service String Advice",
        # conditional). This segment and the UNB segment (Interchange Header)
        # must always be written in ASCII, even if after the BGM the files
        # continues with cyrillic or UTF-16.

        # If it does not exist, return a default.
        if not message[:3] == "UNA":
            return characters

        # Get the character definitions
        chars = message[3:9]
        characters.is_extracted_from_message = True

        characters.component_separator = chars[0]
        characters.data_separator = chars[1]
        characters.decimal_point = chars[2]
        characters.escape_character = chars[3]
        characters.reserved_character = chars[4]
        characters.segment_terminator = chars[5]

        return characters

    def convert_tokens_to_segments(self, tokens: list, characters: Characters):
        """Convert the tokenized message into an array of segments.
        :param with_una: whether the UNA segment should be included
        :param tokens: The tokens that make up the message
        :param characters: the control characters to use
        :type tokens: list of Token
        :rtype list of Segment
        """

        segments = []
        current_segment = []
        data_element = None
        in_segment = False
        empty_component_counter = 0

        for token in tokens:

            # If we're in the middle of a segment, check if we've reached the end
            if in_segment:

                if token.type == Token.Type.TERMINATOR:
                    in_segment = False
                    if len(data_element) == 0:  # empty element
                        data_element = ""
                    if len(data_element) == 1:
                        # use a str instead of a list
                        data_element = data_element[0]

                    current_segment.append(data_element)
                    data_element = []
                    continue

            # If we're not in a segment, then start a new empty one now
            # and add it to the list. Also create a new empty data element,
            # because if the next token is a DATA_SEPARATOR, at least we have
            # an empty string to save into the segment then.
            else:
                current_segment = []
                segments.append(current_segment)
                data_element = []
                in_segment = True

            # Whenever we reach a data separator (+), we add the currently
            # collected data element to the current segment and reset the
            # data_element to an empty list []
            if token.type == Token.Type.DATA_SEPARATOR:
                if len(data_element) == 0:  # empty element
                    data_element = ""
                elif len(data_element) == 1:
                    data_element = data_element[0]

                current_segment.append(data_element)

                data_element = []
                empty_component_counter = 0
                continue

            # Whenever we reach a component data separator (:), we know that
            # the whole data element is a composite, so increment the counter
            # this is especially needed when more than one component data
            # separators are in a row "23:::56"
            if token.type == Token.Type.COMPONENT_SEPARATOR:
                empty_component_counter += 1
                continue

            # here we can be sure that the token value is normal "content"
            # first backfill empty strings for skipped component data (:::)
            for i in range(
                1,
                empty_component_counter
                if data_element
                else empty_component_counter + 1,
            ):
                data_element.append("")

            data_element.append(token.value)
            empty_component_counter = 0
            continue

        self.factory.characters = characters
        for segment in segments:
            name = segment.pop(0)
            yield self.factory.create_segment(name, *segment)



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import copy
import pytest

from pydifact.segmentcollection import SegmentCollection
from pydifact.segments import Segment
from pydifact.serializer import Serializer


@pytest.fixture
def serializer():
    return Serializer()


def assert_segments(serializer, expected: str, segments: list):
    """Helper function add default UNA header and terminator, and compare string to segment"""
    expected = "UNA:+,? '" + expected + "'"
    collection = serializer.serialize(segments, with_una_header=True)
    assert expected == collection


def test_una_integrity1():
    m = SegmentCollection()
    initstring = ":+,? '"
    m.add_segment(Segment("UNA", initstring))
    assert m.serialize() == "UNA" + initstring


def test_UNA_integrity2():
    m = SegmentCollection()
    initstring = ":+.? '"
    m.add_segment(Segment("UNA", initstring))
    assert m.serialize() == "UNA" + initstring


def test_empty_segment_list():
    m = SegmentCollection()
    assert m.serialize() == ""


def test_basic1(serializer):
    assert_segments(serializer, "RFF+PD:50515", [Segment("RFF", ["PD", "50515"])])


def test_basic2(serializer):
    assert_segments(serializer, "RFF+PD+50515", [Segment("RFF", "PD", "50515")])


def test_with_una_in_segments(serializer):
    assert_segments(
        serializer,
        "RFF+PD+45761",
        [Segment("UNA", ":+,? '"), Segment("RFF", "PD", "45761")],
    )


def test_escape_character(serializer):
    assert_segments(
        serializer,
        "ERC+10:The message does not make sense??",
        [Segment("ERC", ["10", "The message does not make sense?"])],
    )


def test_escape_component_separator(serializer):
    assert_segments(
        serializer, "ERC+10:Name?: Craig", [Segment("ERC", ["10", "Name: Craig"])]
    )


def test_escape_data_separator(serializer):
    assert_segments(
        serializer, "DTM+735:?+0000:406", [Segment("DTM", ["735", "+0000", "406"])]
    )


def test_escape_decimal_point(serializer):
    assert_segments(serializer, "QTY+136:12,235", [Segment("QTY", ["136", "12,235"])])


def test_escape_segment_terminator(serializer):
    assert_segments(serializer, "ERC+10:Craig?'s", [Segment("ERC", ["10", "Craig's"])])


def test_escape_sequence(serializer):
    assert_segments(
        serializer,
        "ERC+10:?:?+???' - ?:?+???' - ?:?+???'",
        [Segment("ERC", ["10", ":+?' - :+?' - :+?'"])],
    )


def test_no_mutation(serializer):
    segments1 = [Segment("ERC", [":+?'"])]
    segments2 = copy.deepcopy(segments1)
    serializer.serialize(segments1, with_una_header=True)
    assert segments1 == segments2



# Pydifact - a python edifact library
#
# Copyright (c) 2019 Christian González
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from copy import copy


class Characters:
    """A set of control characters to use."""

    def __init__(self):
        # The control character used to separate components.
        self.component_separator = ":"

        # The control character used to separate data elements.
        self.data_separator = "+"

        # The control character used as a decimal point.
        self.decimal_point = ","

        # The control character used as an escape character.
        self.escape_character = "?"

        # Reserved for future use
        self.reserved_character = " "

        # The control character used as an segment terminator.
        self.segment_terminator = "'"

        self.line_terminators = [" ", "\r", "\n"]

    @classmethod
    def from_str(cls, string: str) -> "Characters":
        """Returns a new instance with control characters set to given string.
        :param string: The string to set the control characters to
        """
        if string[0:3] == "UNA":
            string = string[3:9]
        assert len(string) >= 6

        characters = cls()
        characters.component_separator = string[0]
        characters.data_separator = string[1]
        characters.decimal_point = string[2]
        characters.escape_character = string[3]
        characters.reserved_character = string[4]
        characters.segment_terminator = string[5]
        return characters

    def with_control_character(self, cc_type: str, char: str):
        """Set a control character.

        :param cc_type: The type of control character to set,
            as one of the following attribute strings:
            componentSeparator, dataSeparator, decimalPoint,
            escapeCharacter, segmentTerminator
        :param char: The character to set it to
        :return: clone of self
        """
        if len(char) != 1:
            raise ValueError("control characters must only be a single character")

        # set the attribute dynamically.
        if not hasattr(self, cc_type):
            raise AttributeError(
                "{} doesn't have an attribute with the name '{}'".format(self, cc_type)
            )

        other = copy(self)
        setattr(other, cc_type, char)

        # return clone
        return other

    @property
    def service_string_advice(self) -> str:
        return "UNA" + str(self)

    @property
    def escaped_syntax_dic(self):
        return {
            self.escape_character: self.escape_character + self.escape_character,
            self.component_separator: self.escape_character + self.component_separator,
            self.data_separator: self.escape_character + self.data_separator,
            self.segment_terminator: self.escape_character + self.segment_terminator,
        }

    def __str__(self) -> str:
        return str(
            self.component_separator
            + self.data_separator
            + self.decimal_point
            + self.escape_character
            + self.reserved_character
            + self.segment_terminator
        )

    def __repr__(self):
        return "'{}'".format(self.__str__())

    def __eq__(self, other):
        if type(other) == str:
            other = Characters.from_str(other)
        return (
            (self.component_separator == other.component_separator)
            and (self.data_separator == other.data_separator)
            and (self.decimal_point == other.decimal_point)
            and (self.escape_character == other.escape_character)
            and (self.reserved_character == other.reserved_character)
            and (self.segment_terminator == other.segment_terminator)
        )



# Pydifact - a python edifact library
#
# Copyright (c) 2019 Christian González
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import collections
from typing import Callable, Iterable, List, Optional, Tuple, Union
import datetime
import warnings

from pydifact.api import EDISyntaxError
from pydifact.parser import Parser
from pydifact.segments import Segment
from pydifact.serializer import Serializer
from pydifact.control import Characters
import codecs


class AbstractSegmentsContainer:
    """Represent a collection of EDI Segments for both reading and writing."""

    def __init__(self, extra_header_elements: List[Union[str, List[str]]] = []):
        """
        :param extra_header_elements: a list of elements to be appended at the end
          of the header segment (same format as Segment() constructor *elements).
        """

        # The segments that make up this message
        self.segments = []
        self.characters = Characters()

        self.extra_header_elements = extra_header_elements

        # Flag whether the UNA header is present
        self.has_una_segment = False

    @classmethod
    def from_str(cls, string: str) -> "SegmentCollection":
        """Create a SegmentCollection instance from a string.
        :param string: The EDI content
        """
        segments = Parser().parse(string)

        return cls.from_segments(segments)

    @classmethod
    def from_segments(
        cls, segments: list or collections.Iterable
    ) -> "AbstractSegmentsContainer":
        """Create a new AbstractSegmentsContainer instance from a iterable list of segments.

        :param segments: The segments of the EDI interchange
        :type segments: list/iterable of Segment
        """

        # create a new instance of AbstractSegmentsContainer and return it
        # with the added segments
        return cls().add_segments(segments)

    def get_segments(
        self, name: str, predicate: Callable[[Segment], bool] = None
    ) -> list:
        """Get all the segments that match the requested name.

        :param name: The name of the segments to return
        :param predicate: Optional predicate callable that returns True if the given segment matches a condition
        :rtype: list of Segment
        """
        for segment in self.segments:
            if segment.tag == name and (predicate is None or predicate(segment)):
                yield segment

    def get_segment(
        self, name: str, predicate: Callable[[Segment], bool] = None
    ) -> Optional[Segment]:
        """Get the first segment that matches the requested name.

        :return: The requested segment, or None if not found
        :param name: The name of the segment to return
        :param predicate: Optional predicate that must match on the segments
           to return
        """
        for segment in self.get_segments(name, predicate):
            return segment

        return None

    def split_by(
            self,
            start_segment_tag: str,
    ) -> Iterable['RawSegmentCollection']:
        """Split a segment collection by tag.

        Everything before the first start segment is ignored, so if no matching
        start segment is found at all, returned result is empty.


        :param start_segment_tag:
          the segment tag we want to use as separator

        :return: generator of segment collections. The start tag is included in
          each yielded collection
        """
        current_list = None

        for segment in self.segments:
            if segment.tag == start_segment_tag:
                if current_list:
                    yield current_list
                current_list = RawSegmentCollection.from_segments([segment])
            else:
                if current_list is not None:
                    current_list.add_segment(segment)
                else:
                    continue # we are not yet inside a group
        if current_list is not None:
            yield current_list


    def add_segments(
        self, segments: List[Segment] or collections.Iterable
    ) -> "AbstractSegmentsContainer":
        """Add multiple segments to the collection. Passing a UNA segment means setting/overriding the control
        characters and setting the serializer to output the Service String Advice. If you wish to change the control
        characters from the default and not output the Service String Advice, change self.characters instead,
        without passing a UNA Segment.

        :param segments: The segments to add
        :type segments: list or iterable of Segments
        """
        for segment in segments:
            self.add_segment(segment)

        return self

    def add_segment(self, segment: Segment) -> "AbstractSegmentsContainer":
        """Append a segment to the collection.

        :param segment: The segment to add
        """
        self.segments.append(segment)
        return self

    def get_header_segment(self) -> Optional[Segment]:
        """Craft and return this container header segment (if any)

        :returns: None if there is no header for that container
        """
        return None

    def get_footer_segment(self) -> Optional[Segment]:
        """Craft and return this container footer segment (if any)
        :returns: None if there is no footer for that container
        """
        return None

    def serialize(self, break_lines: bool = False) -> str:
        """Serialize all the segments added to this object.
        :param break_lines: if True, insert line break after each segment terminator.
        """
        header = self.get_header_segment()
        footer = self.get_footer_segment()
        out = []

        if header:
            out.append(header)
        out += self.segments
        if footer:
            out.append(footer)

        return Serializer(self.characters).serialize(
            out,
            self.has_una_segment,
            break_lines,
        )

    def __str__(self) -> str:
        """Allow the object to be serialized by casting to a string."""
        return self.serialize()


class FileSourcableMixin:
    """
    For backward compatibility

    For v0.2 drop this class and move from_file() to Interchange class.
    """

    @classmethod
    def from_file(cls, file: str, encoding: str = "iso8859-1") -> "FileSourcableMixin":
        """Create a Interchange instance from a file.

        Raises FileNotFoundError if filename is not found.
        :param encoding: an optional string which specifies the encoding. Default is "iso8859-1".
        :param file: The full path to a file that contains an EDI message.
        :rtype: FileSourcableMixin
        """
        # codecs.lookup raises an LookupError if given codec was not found:
        codecs.lookup(encoding)

        with open(file, encoding=encoding) as f:
            collection = f.read()
        return cls.from_str(collection)


class UNAHandlingMixin:
    """
    For backward compatibility

    For v0.2 drop this class and move add_segment() to Interchange class.
    """

    def add_segment(self, segment: Segment) -> "UNAHandlingMixin":
        """Append a segment to the collection. Passing a UNA segment means setting/overriding the control
        characters and setting the serializer to output the Service String Advice. If you wish to change the control
        characters from the default and not output the Service String Advice, change self.characters instead,
        without passing a UNA Segment.

        :param segment: The segment to add
        """
        if segment.tag == "UNA":
            self.has_una_segment = True
            self.characters = Characters.from_str(segment.elements[0])
            return self
        return super().add_segment(segment)


class SegmentCollection(
    FileSourcableMixin, UNAHandlingMixin, AbstractSegmentsContainer
):
    """
    For backward compatibility. Drop it in v0.2

    Will be replaced by Interchange or RawSegmentCollection depending on the need.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "SegmentCollection is deprecated and will no longer be available in v0.2, "
            "replace it with Interchange or RawSegmentCollection",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)

    @classmethod
    def from_file(cls, *args, **kwargs) -> "SegmentCollection":
        warnings.warn(
            "SegmentCollection.from_file will be removed in v0.2, "
            "Use Interchange class instead",
            DeprecationWarning,
        )
        return super().from_file(*args, **kwargs)

    def add_segment(self, segment: Segment) -> "SegmentCollection":
        if segment.tag == "UNA":
            warnings.warn(
                "SegmentCollection will be removed in v0.2, "
                "For UNA handling, use Interchange class instead",
                DeprecationWarning,
            )
        return super().add_segment(segment)


class RawSegmentCollection(AbstractSegmentsContainer):
    """
    A way to analyze arbitrary bunch of edifact segments.

    Similar to the deprecated SegmentCollection, but lacking from_file() and UNA support.

    If you are handling an Interchange or a Message, you may want to prefer
    those classes to RawSegmentCollection, as they offer more features and
    checks.
    """

    pass


class Message(AbstractSegmentsContainer):
    """
    A message (started by UNH segment, ended by UNT segment)

    Optional features of UNH are not yet supported.

    https://www.stylusstudio.com/edifact/40100/UNH_.htm
    https://www.stylusstudio.com/edifact/40100/UNT_.htm
    """

    def __init__(self, reference_number: str, identifier: Tuple, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reference_number = reference_number
        self.identifier = identifier

    @property
    def type(self) -> str:
        return self.identifier[0]

    @property
    def version(self) -> str:
        """
        Gives version number and release number.

        :return: message version, parsable by pkg_resources.parse_version()
        """
        return "{}.{}".format(self.identifier[1], self.identifier[2])

    def get_header_segment(self) -> Segment:
        return Segment(
            "UNH",
            self.reference_number,
            [str(i) for i in self.identifier],
            *self.extra_header_elements,
        )

    def get_footer_segment(self) -> Segment:
        return Segment(
            "UNT",
            self.reference_number,
            str(len(self.segments)),
        )


class Interchange(FileSourcableMixin, UNAHandlingMixin, AbstractSegmentsContainer):
    """
    An interchange (started by UNB segment, ended by UNZ segment)

    Optional features of UNB are not yet supported.

    Functional groups are not yet supported

    Messages are supported, see get_message() and get_message(), but are
    optional: interchange segments can be accessed without going through
    messages.

    https://www.stylusstudio.com/edifact/40100/UNB_.htm
    https://www.stylusstudio.com/edifact/40100/UNZ_.htm
    """

    def __init__(
        self,
        sender: str,
        recipient: str,
        control_reference: str,
        syntax_identifier: Tuple[str, int],
        delimiters: Characters = Characters(),
        timestamp: datetime.datetime = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.sender = sender
        self.recipient = recipient
        self.control_reference = control_reference
        self.syntax_identifier = syntax_identifier
        self.delimiters = delimiters
        self.timestamp = timestamp or datetime.datetime.now()

    def get_header_segment(self) -> Segment:
        return Segment(
            "UNB",
            [str(i) for i in self.syntax_identifier],
            self.sender,
            self.recipient,
            ["{:%y%m%d}".format(self.timestamp), "{:%H%M}".format(self.timestamp)],
            self.control_reference,
            *self.extra_header_elements,
        )

    def get_footer_segment(self) -> Segment:
        return Segment(
            "UNZ",
            str(len(self.segments)),
            self.control_reference,
        )

    def get_messages(self) -> List[Message]:
        message = None
        for segment in self.segments:
            if segment.tag == "UNH":
                if not message:
                    message = Message(segment.elements[0], segment.elements[1])
                else:
                    raise EDISyntaxError(
                        "Missing UNT segment before new UNH: {}".format(segment)
                    )
            elif segment.tag == "UNT":
                if message:
                    yield message
                else:
                    raise EDISyntaxError(
                        'UNT segment without matching UNH: "{}"'.format(segment)
                    )
            else:
                if message:
                    message.add_segment(segment)

    def add_message(self, message: Message) -> "Interchange":
        segments = (
            [message.get_header_segment()]
            + message.segments
            + [message.get_footer_segment()]
        )
        self.add_segments(i for i in segments if i is not None)
        return self

    @classmethod
    def from_segments(
        cls, segments: Union[list, collections.Iterable]
    ) -> "Interchange":
        segments = iter(segments)

        first_segment = next(segments)
        if first_segment.tag == "UNA":
            unb = next(segments)
        elif first_segment.tag == "UNB":
            unb = first_segment
        else:
            raise EDISyntaxError("An interchange must start with UNB or UNA and UNB")
        # Loosy syntax check :
        if len(unb.elements) < 4:
            raise EDISyntaxError("Missing elements in UNB header")

        datetime_str = "-".join(unb.elements[3])
        timestamp = datetime.datetime.strptime(datetime_str, "%y%m%d-%H%M")
        interchange = Interchange(
            syntax_identifier=unb.elements[0],
            sender=unb.elements[1],
            recipient=unb.elements[2],
            timestamp=timestamp,
            control_reference=unb.elements[4],
        )

        if first_segment.tag == "UNA":
            interchange.has_una_segment = True
            interchange.characters = Characters.from_str(first_segment.elements[0])

        return interchange.add_segments(
            segment for segment in segments if segment.tag != "UNZ"
        )



"""
This file contains common used classes described in the EDIFACT syntax.

There are multiple versions (1-4) of the EDIFACT standard, and each differ slightly.
As one project will probably never use more than one syntax version simultaneously,
it is splitted into 4 different python files, each which could be imported using

    from pydifact.syntax.v[1..4] import ...

herby providing the same structure.
"""

import re

from pydifact import Segment, Characters


def assert_a(s, length):
    """checks if s consists of only alphabetical characters, and has a given length."""
    assert str(s).isalpha()
    assert len(s) == length


def assert_a_max(s, length):
    """checks if s consists of only alphabetical characters, and has a given maximum length."""
    assert str(s).isalpha() or s == ""
    assert len(s) <= length


def assert_n(s, length):
    """checks if s is numeric and has a given length."""
    assert int(s)
    assert len(s) == length


def assert_n_max(s, length):
    """checks if s is numeric and has a given length."""
    assert int(s)
    assert len(s) <= length


def assert_an(s, length):
    """checks if s is alphanumeric and has a given length."""
    assert len(s) == length


def assert_an_max(s, length):
    """checks if s is alphanumeric and has a given length."""
    assert len(s) <= length


def assert_a(s, length):
    """checks if s only contains characters (no numbers) and has a given length."""
    assert len(s) == length
    # TODO only alpha


def assert_a_max(s, length):
    """checks if s only contains characters (no numbers) and has a given length."""
    assert len(s) <= length


def assert_format(s, fmt_str):
    assert re.match(fmt_str, s)


class UNASegment(Segment):
    """Service String Advice."""

    tag = "UNA"

    def __init__(self, characters: Characters or str = None):
        if not characters:
            characters = Characters()
        assert_an(str(characters), 6)
        super().__init__("UNA", characters)



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
from pydifact.parser import Parser
from pydifact.segments import Segment

import pytest

# @pytest.fixture
# def mocked_tokenizer(mocker):
#     tokenizer = Tokenizer()
#     mocker.patch("pydifact.tokenizer.Tokenizer")
#     return tokenizer


@pytest.fixture
def parser():
    return Parser()


@pytest.fixture
def default_una_segment():
    return Segment("UNA", ":+,? '")


# def get_control_characters(mocker, parser, collection: str, tokenizer=None) -> Characters:
#     """Returns the control characters from the given collection.
#
#     :return: the collection without the "UNA123456" header string
#     """
#
#     if tokenizer is None:
#         mocker.patch("pydifact.tokenizer.Tokenizer")
#         tokenizer = Tokenizer()
#
#         tokenizer.set_control_character.assert_called_once_with(
#             "set_component_separator", 1
#         )
#         tokenizer.set_control_character.assert_called_once_with(
#             "set_data_separator", 2
#         )
#         tokenizer.set_control_character.assert_called_once_with(
#             "set_decimal_point", 3
#         )
#         tokenizer.set_control_character.assert_called_once_with(
#             "set_escape_character", 4
#         )
#         tokenizer.set_control_character.assert_called_once_with(
#             "set_segment_terminator", 6
#         )
#     # FIXME: use characters
#     return parser.get_control_characters(collection, tokenizer)

#    def test_setup_special_characters1(self):
#
#        tokenizer = mock.MagicMock(spec=Tokenizer)
#        collection = self.get_control_characters("TEST", tokenizer)
#        self.assertEqual("TEST", collection)


def test_setup_special_characters_only(parser):
    assert parser.get_control_characters("UNA123456") == "123456"


def test_setup_special_characters_with_message(parser):
    assert parser.get_control_characters("UNA123456TEST") == "123456"


def test_setup_special_characters_with_linefeed(parser):
    assert parser.get_control_characters("UNA123456\nTEST") == "123456"


def test_setup_special_characters_with_crlf(parser):
    assert parser.get_control_characters("UNA123456\r\nTEST") == "123456"


def _assert_segments(parser, default_una_segment, collection: str, segments: list):
    """This function asserts that the given collection, when parsed with
    Parser.parse(), produces exactly the list output given by segments.
    :param collection: The collection to parse. The UNA string is added.
    :param segments: The expected segments list
    """

    input_str = "UNA:+,? '\n" + collection + "'\n"
    result = list(parser.parse(input_str))
    # print("input segments: {}".format(segments[0]))
    # print("parser result:  {}".format(result[0]))
    assert segments == result


def test_compare_equal_segments(parser, default_una_segment):
    """Just make sure that comparing Segment objects works"""
    a = Segment("RFF", ["PD", "50515"])
    b = Segment("RFF", ["PD", "50515"])
    assert a == b
    assert (
        a is not b
    ), "Two separatedly, but visually identically created Segment objects may not be the same object."


def test_una_parser1(parser):
    # UNA headers are a special parsing task and must be processed correctly.
    tokens = parser.parse("UNA:+,? 'TEST'")
    assert next(tokens) == Segment("TEST")


def test_una_parser2(parser):
    # UNA headers are a special parsing task and must be processed correctly.
    tokens = parser.parse("UNA123456TEST6")
    assert next(tokens) == Segment("TEST")


def test_una_parser3(parser):
    # UNA headers are a special parsing task and must be processed correctly.
    tokens = parser.parse("UNA12345'TEST'")
    assert next(tokens) == Segment("TEST")


def test_basic1(parser, default_una_segment):
    _assert_segments(
        parser, default_una_segment, "RFF+PD:50515", [Segment("RFF", ["PD", "50515"])]
    )


def test_basic2(parser, default_una_segment):

    _assert_segments(
        parser, default_una_segment, "RFF+PD+50515", [Segment("RFF", "PD", "50515")]
    )


def test_escape_character(parser, default_una_segment):
    _assert_segments(
        parser,
        default_una_segment,
        "ERC+10:The message does not make sense??",
        [Segment("ERC", ["10", "The message does not make sense?"])],
    )


def test_escape_component_separator(parser, default_una_segment):

    _assert_segments(
        parser,
        default_una_segment,
        "ERC+10:Name?: Craig",
        [Segment("ERC", ["10", "Name: Craig"])],
    )


def test_escape_data_separator(parser, default_una_segment):

    _assert_segments(
        parser,
        default_una_segment,
        "DTM+735:?+0000:406",
        [Segment("DTM", ["735", "+0000", "406"])],
    )


def test_escape_decimal_point(parser, default_una_segment):

    _assert_segments(
        parser,
        default_una_segment,
        "QTY+136:12,235",
        [Segment("QTY", ["136", "12,235"])],
    )


def test_escape_segment_terminator(parser, default_una_segment):

    _assert_segments(
        parser,
        default_una_segment,
        "ERC+10:Craig?'s",
        [Segment("ERC", ["10", "Craig's"])],
    )


def test_escape_sequence(parser, default_una_segment):

    _assert_segments(
        parser,
        default_una_segment,
        "ERC+10:?:?+???' - ?:?+???' - ?:?+???'",
        [Segment("ERC", ["10", ":+?' - :+?' - :+?'"])],
    )


def test_compound_starts_with_skipped(parser, default_una_segment):

    _assert_segments(
        parser, default_una_segment, "IMD+::A", [Segment("IMD", ["", "", "A"])]
    )


def test_compound_contains_one_skipped(parser, default_una_segment):

    _assert_segments(
        parser, default_una_segment, "IMD+A::B", [Segment("IMD", ["A", "", "B"])]
    )


def test_compound_contains_two_skipped(parser, default_una_segment):

    _assert_segments(
        parser, default_una_segment, "IMD+A:::B", [Segment("IMD", ["A", "", "", "B"])]
    )



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest
from pydifact.token import Token


class TokenTest(unittest.TestCase):
    def test_type(self):
        token = Token(Token.Type.CONTENT, "ok")
        self.assertEqual(Token.Type.CONTENT, token.type)

    def test_value(self):
        token = Token(Token.Type.CONTENT, "ok")
        self.assertEqual("ok", token.value)


if __name__ == "__main__":
    unittest.main()



class PluginMount(type):
    """Generic plugin mount point (= entry point) for pydifact plugins.

    .. note::
        Plugins that have an **__omitted__** attriute are not added to the list!
    """

    # thanks to Marty Alchin!

    def __init__(cls, name, bases, attrs):
        if not hasattr(cls, "plugins"):
            cls.plugins = []
        else:
            if not hasattr(cls, "__omitted__"):
                cls.plugins.append(cls)


class EDISyntaxError(SyntaxError):
    """A Syntax error within the parsed EDIFACT file was found."""



# Pydifact - a python edifact library
#
# Copyright (c) 2019 Christian González
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from pydifact import segmentcollection, parser, segments, serializer, token, tokenizer

__version__ = "0.1.1"

from .control.characters import Characters
from .segmentcollection import SegmentCollection
from .parser import Parser
from .segments import Segment
from .serializer import Serializer
from .token import Token
from .tokenizer import Tokenizer

# up-to-date information for newest standard
# https://www.unece.org/tradewelcome/un-centre-for-trade-facilitation-and-e-business-uncefact/outputs/standards/unedifact/directories/2011-present.html

# Joint Working Group 1 (JWG 1)
# https://www.gefeg.com/jswg/

# EDIFACT Syntax Rules
# http://www.unece.org/fileadmin/DAM/trade/edifact/untdid/d422_s.htm#normative

# UNECE Syntax Implementation Guidelines
# https://www.unece.org/fileadmin/DAM/trade/untdid/texts/d423.htm

# UNECE Message Design Guidelines
# https://www.unece.org/fileadmin/DAM/trade/untdid/texts/d424_d.htm

# generic EDIFACT implementation tutorial
# http://www.gxs.co.uk/wp-content/uploads/tutorial_edifact.pdf

# https://ecosio.com/de/blog/aufbau-einer-edifact-datei/ (German)

# https://www.hcs.at/wp-content/uploads/2012/12/hcs_AEK_Edifact_Text.pdf (German)
# https://www.hcs.at/wp-content/uploads/2012/12/hcs_AEK_Edifact_Labor.pdf (German)



#    pydifact - a python edifact library
#    Copyright (C) 2017-2018  Christian González
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest

from pydifact.control import Characters


class TestControlCharacters(unittest.TestCase):
    def setUp(self):
        self.cc = Characters()

    def test_wrong_attribute(self):
        self.assertRaises(
            AttributeError, self.cc.with_control_character, "wrongtype", "+"
        )

    def test_wrong_character_size(self):
        self.assertRaises(
            ValueError, self.cc.with_control_character, "decimal_point", ",."
        )

    def test_correct_parameters(self):

        d = self.cc.with_control_character("component_separator", "/")

        self.assertEqual(
            self.cc.with_control_character(
                "component_separator", "/"
            ).component_separator,
            "/",
        )

        self.assertEqual(
            self.cc.with_control_character("data_separator", "/").data_separator, "/"
        )

        self.assertEqual(
            self.cc.with_control_character("decimal_point", "/").decimal_point, "/"
        )

        self.assertEqual(
            self.cc.with_control_character("escape_character", "/").escape_character,
            "/",
        )

        self.assertEqual(
            self.cc.with_control_character(
                "reserved_character", "/"
            ).reserved_character,
            "/",
        )

        self.assertEqual(
            self.cc.with_control_character(
                "segment_terminator", "/"
            ).segment_terminator,
            "/",
        )



# Pydifact - a python edifact library
#
# Copyright (c) 2019 Christian González
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from pydifact.token import Token
from pydifact.control.characters import Characters
from typing import List, Optional


class Tokenizer:
    """Convert EDI messages into tokens for parsing."""

    def __init__(self):
        super().__init__()

        # The message that we are tokenizing.
        self._message = []

        self._current_chars = []

        # The current character from the message we are dealing with.
        self._char = ""

        # The stored characters for the next token.
        self._string = ""

        # bool isEscaped If the current character has been escaped.
        self.isEscaped = False

        # The control characters for the message
        self.characters = None

        self.token_selector = {}

        self._message_index = 0

    def get_tokens(self, message: str, characters: Characters = None) -> List[Token]:
        """Convert the passed message into tokens.
        :param characters: the Control Characters to use for tokenizing. If omitted, use a default set.
        :param message: The EDI message
        :return: Token[]
        """

        self.characters = characters or Characters()
        self._char = None
        self._message = iter(message)
        self._message_index = 0
        self.read_next_char()

        self.token_selector = {
            self.characters.component_separator: Token.Type.COMPONENT_SEPARATOR,
            self.characters.data_separator: Token.Type.DATA_SEPARATOR,
            self.characters.segment_terminator: Token.Type.TERMINATOR,
        }

        while not self.end_of_message():
            yield self.get_next_token()

    def read_next_char(self) -> None:
        """Read the next character from the message.

        If the character is an escape character, set the isEscaped flag to
        True, get the one after it and store that character in the internal storage."""

        # first, get the next char.
        self._char = self.get_next_char()

        # If the last character was escaped, this one can't possibly be an escaped one.
        if self.isEscaped:
            self.isEscaped = False

        # If this is the escape character, then read the next one, store it and
        # flag it as escaped
        if self._char == self.characters.escape_character:
            self.isEscaped = True
            self._char = self.get_next_char()

    def get_next_char(self) -> Optional[str]:
        """Get the next character from the message."""
        try:
            return next(self._message)
        except StopIteration:
            return

    def get_next_token(self) -> Optional[Token]:
        """Get the next token from the message."""

        # If we're not escaping this character then see if it's
        # a control character

        token_type = not self.isEscaped and self.token_selector.get(self._char)
        if token_type:
            self.store_current_char_and_read_next()
            token = Token(token_type, self.extract_stored_chars())
            if token_type == Token.Type.TERMINATOR:
                while self._char in self.characters.line_terminators:
                    self.read_next_char()
            return token

        while not self.is_control_character():
            if self.end_of_message():
                raise RuntimeError("Unexpected end of EDI message")

            self.store_current_char_and_read_next()
        return Token(Token.Type.CONTENT, self.extract_stored_chars())

    def is_control_character(self) -> bool:
        """Check if the current character is a control character."""

        if self.isEscaped:
            return False

        return self._char in self.token_selector

    def store_current_char_and_read_next(self) -> None:
        """Store the current character and read the
        next one from the message."""

        self._current_chars.append(self._char)
        self.read_next_char()

    def extract_stored_chars(self) -> str:
        """Return the previously stored characters and empty the store."""

        chars = self._current_chars
        self._current_chars = []
        return "".join(chars)

    def end_of_message(self) -> bool:
        """Check if we've reached the end of the message"""
        return self._char is None

    def __str__(self):
        return "".join(self._current_chars)
